/*
 * Botendo
 * Copyright (C) 2023 moonleay
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

package net.moonleay.botendo

import com.kotlindiscord.kord.extensions.ExtensibleBot
import dev.kord.common.Color
import dev.kord.common.entity.Snowflake
import dev.kord.core.behavior.getChannelOf
import dev.kord.core.behavior.interaction.response.respond
import dev.kord.core.entity.channel.VoiceChannel
import dev.kord.core.event.interaction.ButtonInteractionCreateEvent
import dev.kord.core.event.user.VoiceStateUpdateEvent
import dev.kord.core.on
import dev.kord.rest.builder.message.embed
import dev.schlaubi.lavakord.LavaKord
import dev.schlaubi.lavakord.audio.Link
import dev.schlaubi.lavakord.kord.getLink
import dev.schlaubi.lavakord.kord.lavakord
import kotlinx.coroutines.flow.count
import net.moonleay.botendo.data.CredentialManager
import net.moonleay.botendo.extensions.music.*
import net.moonleay.botendo.extensions.music.components.MusicManager
import net.moonleay.botendo.extensions.util.InfoExtension
import net.moonleay.botendo.util.Logger
import net.moonleay.botendo.util.MessageUtil

object Bot {
    //The kord object gets set at app launch
    private lateinit var bot: ExtensibleBot
    lateinit var lava: LavaKord
    suspend fun start() {
        Logger.out("Starting Bot...")

        //Load config
        CredentialManager.load()

        //Don't run the bot when there is no bot token in config
        if (CredentialManager.token == "empty") {
            Logger.out("The config does not contain a bot token")
            return
        }

        //Add Bot token to kord
        //kord = Kord(CredentialManager.token)
        bot = ExtensibleBot(CredentialManager.token) {
            applicationCommands {
                enabled = true
            }

            extensions {
                add(::InfoExtension)
                add(::PlayExtension)
                add(::StopExtension)
                add(::SkipExtension)
                add(::NowPlayingExtension)
                add(::QueueExtension)
                add(::UpsertExtension)
                add(::SeekExtension)
            }

            this.presence {
                this.streaming("music", "https://twitch.tv/moonleaytv")
            }
        }

        //Add LavaLink
        lava = bot.kordRef.lavakord()

        //Add the LavaLink node from config
        lava.addNode(CredentialManager.linkip, CredentialManager.linkpw)

        //Register Button presses
        bot.kordRef.on<ButtonInteractionCreateEvent> {
            val inter = this.interaction
            val response = inter.deferPublicResponse()
            val u = inter.user
            val g = this.interaction.getOriginalInteractionResponse().getGuild()
            for (b in net.moonleay.botendo.buttons.ButtonManager.buttons) {
                if (b.id != inter.componentId)
                    continue
                b.onInteraction(response, g, u)
                return@on
            }
            // Button not found
            response.respond {
                this.embed {
                    this.color = Color(0xE0311A)
                    this.title = "Error"
                    this.description = "Could not find button with id \"${inter.componentId}\".\nPlease report this."
                    this.footer {
                        this.text = MessageUtil.getFooter(u)
                    }
                }
            }
        }

        // Allow the bot to leave the VC if there are no users in the VC
        bot.kordRef.on<VoiceStateUpdateEvent> {
            if (this.state.userId == bot.kordRef.selfId) {
                // Ignore self
                return@on
            }
            val g = this.state.getGuild().asGuild()
            val link = lava.getLink(g.id)
            if (link.state != Link.State.CONNECTED) {
                // Ignore if not connected
                return@on
            }
            val myChannel = link.lastChannelId!!
            if (g.getChannelOrNull(Snowflake(myChannel)) == null) {
                // Ignore if channel cannot be found
                return@on
            }
            val ch = g.getChannelOf<VoiceChannel>(Snowflake(myChannel))
            if (ch.voiceStates.count() > 1) {
                // Ignore if there are other users in the channel
                return@on
            }
            // Stop playing if there are no users in the channel
            val pl = MusicManager.getMusicManager(g)!!.pl
            pl.stopTrack()
            link.destroy()
        }

        //Start the bot
        bot.start()
    }
}
